@@ -1,6 +1,8 @@
-# $Id: HTTP.pm 327 2008-05-28 17:33:06Z rcaputo $
-
 package POE::Component::Client::HTTP;
+BEGIN {
+  $POE::Component::Client::HTTP::VERSION = '0.943';
+}
+# vim: ts=2 sw=2 expandtab
 
 # {{{ INIT
 
@@ -10,12 +12,10 @@ use strict;
 use constant DEBUG      => 0;
 use constant DEBUG_DATA => 0;
 
-use vars qw($VERSION);
-$VERSION = '0.84';
-
 use Carp qw(croak);
 use HTTP::Response;
 use Net::HTTP::Methods;
+use Socket qw(sockaddr_in inet_ntoa);
 
 use POE::Component::Client::HTTP::RequestFactory;
 use POE::Component::Client::HTTP::Request qw(:states :fields);
@@ -30,6 +30,8 @@ BEGIN {
   };
 }
 
+use Socket::GetAddrInfo qw(:newapi getnameinfo NI_NUMERICHOST NI_NUMERICSERV);
+
 use POE qw(
   Driver::SysRW Filter::Stream
   Filter::HTTPHead Filter::HTTPChunk
@@ -168,7 +170,7 @@ sub _poco_weeble_start {
   # have to do this here because it wants a current_session
   $heap->{cm} = POE::Component::Client::Keepalive->new(
     timeout => $heap->{factory}->timeout,
-    $heap->{bind_addr} ? (bind_address => $heap->{bind_addr}) : (),
+    ($heap->{bind_addr} ? (bind_address => $heap->{bind_addr}) : ()),
   ) unless ($heap->{cm});
 }
 
@@ -182,6 +184,7 @@ sub _poco_weeble_stop {
 
   foreach my $request_rec (values %$request) {
     $request_rec->remove_timeout();
+    delete $heap->{ext_request_to_int_id}->{$request_rec->[REQ_HTTP_REQUEST]};
   }
 
   DEBUG and warn "Client::HTTP (alias=$heap->{alias}) stopped.";
@@ -193,7 +196,7 @@ sub _poco_weeble_stop {
 sub _poco_weeble_pending_requests_count {
   my ($heap) = $_[HEAP];
   my $r = $heap->{request} || {};
-  return keys %$r;
+  return scalar keys %$r;
 }
 
 # }}} _poco_weeble_pending_requests_count
@@ -263,7 +266,7 @@ sub _poco_weeble_request {
     $proxy_override, $sender
   );
   $heap->{request}->{$request->ID} = $request;
-  $heap->{request_to_id}->{$http_request} = $request->ID;
+  $heap->{ext_request_to_int_id}->{$http_request} = $request->ID;
 
   my @timeout;
   if ($heap->{factory}->timeout()) {
@@ -272,22 +275,27 @@ sub _poco_weeble_request {
     );
   }
 
-  my $cm_req_id;
   eval {
-      # get a connection from Client::Keepalive
-      $request->[REQ_CONN_ID] = $heap->{cm}->allocate(
-        scheme  => $request->scheme,
-        addr    => $request->host,
-        port    => $request->port,
-        context => $request->ID,
-        event   => 'got_connect_done',
-        @timeout,
-      );
+    # get a connection from Client::Keepalive
+    #
+    # TODO CONNECT - We must ask PCC::Keepalive to establish an http
+    # socket, not https.  The initial proxy interactin is plaintext?
+
+    $request->[REQ_CONN_ID] = $heap->{cm}->allocate(
+      scheme  => $request->scheme,
+      addr    => $request->host,
+      port    => $request->port,
+      context => $request->ID,
+      event   => 'got_connect_done',
+      @timeout,
+    );
   };
   if ($@) {
-      delete $heap->{request}->{$request->ID};
-      # we can reach here for things like host being invalid.
-      $request->error(400, $@);
+    delete $heap->{request}->{$request->ID};
+    delete $heap->{ext_request_to_int_id}->{$http_request};
+
+    # we can reach here for things like host being invalid.
+    $request->error(400, $@);
   }
 }
 
@@ -302,10 +310,24 @@ sub _poco_weeble_connect_done {
   my $connection = $response->{'connection'};
   my $request_id = $response->{'context'};
 
+  # Can't handle connections if we're shut down.
+  # TODO - How do we still get these?  Were they previously queued or
+  # something?
+  if ($heap->{is_shut_down}) {
+    _internal_cancel(
+      $heap, $request_id, 408, "Request timed out (request canceled)"
+    );
+    return;
+  }
+
   if (defined $connection) {
     DEBUG and warn "CON: request $request_id connected ok...";
 
     my $request = $heap->{request}->{$request_id};
+    unless (defined $request) {
+      DEBUG and warn "CON: ignoring connection for canceled request";
+      return;
+    }
 
     my $block_size = $heap->{factory}->block_size;
 
@@ -325,6 +347,32 @@ sub _poco_weeble_connect_done {
     $heap->{wheel_to_request}->{ $new_wheel->ID() } = $request_id;
 
     $request->[REQ_CONNECTION] = $connection;
+
+    # SSLify needs us to call it's function to get the "real" socket
+    my $peer_addr;
+    if ( $request->scheme eq 'http' ) {
+      $peer_addr = getpeername($new_wheel->get_input_handle());
+    } else {
+      my $socket = $new_wheel->get_input_handle();
+      $peer_addr = getpeername(POE::Component::SSLify::SSLify_GetSocket($socket));
+    }
+
+    if (defined $peer_addr) {
+      my ($error, $address, $port) = getnameinfo(
+        $peer_addr, NI_NUMERICHOST | NI_NUMERICSERV
+      );
+
+      if ($error) {
+        $request->[REQ_PEERNAME] = "error: $error";
+      }
+      else {
+        $request->[REQ_PEERNAME] = "$address.$port";
+      }
+    }
+    else {
+      $request->[REQ_PEERNAME] = "error: $!";
+    }
+
     $request->create_timer($heap->{factory}->timeout);
     $request->send_to_wheel;
   }
@@ -347,9 +395,10 @@ sub _poco_weeble_connect_done {
     DEBUG and warn "I/O: removing request $request_id";
     my $request = delete $heap->{request}->{$request_id};
     $request->remove_timeout();
+    delete $heap->{ext_request_to_int_id}->{$request->[REQ_HTTP_REQUEST]};
 
     # Post an error response back to the requesting session.
-    $request->connect_error("$operation error $errnum: $errstr");
+    $request->connect_error($operation, $errnum, $errstr);
   }
 }
 
@@ -376,6 +425,7 @@ sub _poco_weeble_timeout {
 
   DEBUG and warn "T/O: request $request_id has timer ", $request->timer;
   $request->remove_timeout();
+  delete $heap->{ext_request_to_int_id}->{$request->[REQ_HTTP_REQUEST]};
 
   # There's a wheel attached to the request.  Shut it down.
   if (defined(my $wheel = $request->wheel())) {
@@ -385,9 +435,9 @@ sub _poco_weeble_timeout {
     # Shut down the connection so it's not reused.
     $wheel->shutdown_input();
     delete $heap->{wheel_to_request}->{$wheel_id};
-    delete $heap->{request_to_id}->{$request->[REQ_REQUEST]};
   }
 
+
   DEBUG and do {
     die( "T/O: request $request_id is unexpectedly zero" )
       unless $request->[REQ_STATE];
@@ -399,7 +449,7 @@ sub _poco_weeble_timeout {
 
     # Well, we have a response.  Isn't that nice?  Let's send it.
     if ($request->[REQ_STATE] & (RS_IN_CONTENT | RS_DONE)) {
-      _finish_request($heap, $request, 0);
+      _finish_request($heap, $request);
       return;
     }
 
@@ -430,10 +480,10 @@ sub _poco_weeble_io_flushed {
   );
 
   my $request = $heap->{request}->{$request_id};
-  
+
   # Read content to send from a callback
-  if ( ref $request->[REQ_REQUEST]->content() eq 'CODE' ) {
-    my $callback = $request->[REQ_REQUEST]->content();
+  if ( ref $request->[REQ_HTTP_REQUEST]->content() eq 'CODE' ) {
+    my $callback = $request->[REQ_HTTP_REQUEST]->content();
 
     my $buf = eval { $callback->() };
 
@@ -449,9 +499,10 @@ sub _poco_weeble_io_flushed {
       return;
     }
   }
-  
+
   $request->[REQ_STATE] ^= RS_SENDING;
   $request->[REQ_STATE] = RS_IN_HEAD;
+
   # XXX - Removed a second time.  The first time was in version 0.53,
   # because the EOF generated by shutdown_output() causes some servers
   # to disconnect rather than send their responses.
@@ -466,86 +517,103 @@ sub _poco_weeble_io_error {
   my ($kernel, $heap, $operation, $errnum, $errstr, $wheel_id) =
     @_[KERNEL, HEAP, ARG0..ARG3];
 
-  DEBUG and
-    warn "I/O: wheel $wheel_id encountered $operation error $errnum: $errstr";
+  DEBUG and warn(
+    "I/O: wheel $wheel_id encountered $operation error $errnum: $errstr"
+  );
 
   # Drop the wheel.
   my $request_id = delete $heap->{wheel_to_request}->{$wheel_id};
-  #K or die "!!!: unexpectedly undefined request ID" unless defined $request_id;
 
-  if ($request_id) {
+  # There was no corresponding request?  Nothing left to do here.
+  # We might have got here because the server sent EOF after we were done processing
+  # the request, and deleted it from our cache. ( notes for RT#50231 )
+  return unless $request_id;
 
-    DEBUG and warn "I/O: removing request $request_id";
-    my $request = delete $heap->{request}->{$request_id};
-    delete $heap->{request_to_id}{$request->[REQ_REQUEST]};
-    $request->remove_timeout;
+  DEBUG and warn "I/O: removing request $request_id";
+  my $request = delete $heap->{request}->{$request_id};
+  $request->remove_timeout;
+  delete $heap->{ext_request_to_int_id}{$request->[REQ_HTTP_REQUEST]};
 
-    # Otherwise the remote end simply closed.  If we've got a
-    # pending response, then post it back to the client.
-    DEBUG and warn "STATE is ", $request->[REQ_STATE];
+  # Otherwise the remote end simply closed.  If we've got a pending
+  # response, then post it back to the client.
+  DEBUG and warn "STATE is ", $request->[REQ_STATE];
 
-    # except when we're redirected
-    return if ($request->[REQ_STATE] & RS_REDIRECTED);
+  # Except when we're redirected.  In this case, the connection was but
+  # one step towards our destination.
+  return if ($request->[REQ_STATE] & RS_REDIRECTED);
 
-    # If there was a non-zero error, then something bad happened.  Post
-    # an error response back, if we haven't posted anything before.
-    if ($errnum) {
-      unless ($request->[REQ_STATE] & RS_POSTED) {
-        $request->error(400, "$operation error $errnum: $errstr");
-      }
+  # If there was a non-zero error, then something bad happened.  Post
+  # an error response back, if we haven't posted anything before.
+  if ($errnum) {
+    if ($operation eq "connect") {
+      $request->connect_error($operation, $errnum, $errstr);
       return;
     }
 
-    if (
-      $request->[REQ_STATE] & (RS_IN_CONTENT | RS_DONE) and
-      not $request->[REQ_STATE] & RS_POSTED
-    ) {
-      _finish_request($heap, $request, 0);
-      return;
+    unless ($request->[REQ_STATE] & RS_POSTED) {
+      $request->error(400, "$operation error $errnum: $errstr");
     }
-    elsif ($request->[REQ_STATE] & RS_POSTED) {
-      DEBUG and warn "I/O: Disconnect, remote keepalive timeout or HTTP/1.0.";
-      return;
-    }
-    elsif (not defined $request->[REQ_RESPONSE]) {
-      # never got a response, check for pending data indicating
-      # a LF-free HTTP 0.9 response
-      my $lines = $request->wheel->get_input_filter()->get_pending();
-      my $text = join '' => @$lines;
-      DEBUG and warn "Got ", length($text), " bytes of data without LF.";
-      if ($text =~ /\S/) {
-        # generate response
-        DEBUG and warn(
-          "Generating HTTP response for HTTP/0.9 response without LF."
-        );
-        $request->[REQ_RESPONSE] = HTTP::Response->new(
-          200, 'OK', [ 'Content-Type' => 'text/html' ], $text
-        );
-        $request->[REQ_RESPONSE]->protocol('HTTP/0.9');
-        $request->[REQ_RESPONSE]->request($request->[REQ_REQUEST]);
-        $request->[REQ_STATE] = RS_DONE;
-        $request->return_response;
-        return;
-      } else {
-        unless ($request->[REQ_STATE] & RS_POSTED) {
-          $request->error(400, "incomplete response $request_id");
-          return;
-        }
-      }
-    }
-
-    # We haven't built a proper response.  Send back an error.
-    # Changed to 406 after considering rt.cpan.org 20975.
-    #
-    # 10.4.7 406 Not Acceptable
-    #
-    # The resource identified by the request is only capable of
-    # generating response entities which have content characteristics
-    # not acceptable according to the accept headers sent in the
-    # request.
-
-    $request->error (406, "Response larger than MaxSize - $request_id");
+    return;
   }
+
+  # We seem to have finished with the request.  Send back a response.
+  if (
+    $request->[REQ_STATE] & (RS_IN_CONTENT | RS_DONE) and
+    not $request->[REQ_STATE] & RS_POSTED
+  ) {
+    _finish_request($heap, $request);
+    return;
+  }
+
+  # We have already posted a response, so this is a remote keepalive
+  # timeout or other delayed socket shutdown.  Nothing left to do.
+  if ($request->[REQ_STATE] & RS_POSTED) {
+    DEBUG and warn "I/O: Disconnect, remote keepalive timeout or HTTP/1.0.";
+    return;
+  }
+
+  # We never received a response.
+  if (not defined $request->[REQ_RESPONSE]) {
+    # Check for pending data indicating a LF-free HTTP 0.9 response.
+    my $lines = $request->wheel->get_input_filter()->get_pending();
+    my $text = join '' => @$lines;
+    DEBUG and warn "Got ", length($text), " bytes of data without LF.";
+
+    # If we have data, build and return a response from it.
+    if ($text =~ /\S/) {
+      DEBUG and warn(
+        "Generating HTTP response for HTTP/0.9 response without LF."
+      );
+      $request->[REQ_RESPONSE] = HTTP::Response->new(
+        200, 'OK', [
+          'Content-Type'  => 'text/html',
+          'X-PCCH-Peer'   => $request->[REQ_PEERNAME],
+        ], $text
+      );
+      $request->[REQ_RESPONSE]->protocol('HTTP/0.9');
+      $request->[REQ_RESPONSE]->request($request->[REQ_HTTP_REQUEST]);
+      $request->[REQ_STATE] = RS_DONE;
+      $request->return_response;
+      return;
+    }
+
+    # No data received.  This is an incomplete response.
+    $request->error(400, "Incomplete response - $request_id");
+    return;
+  }
+
+  # We haven't built a proper response, and nothing returned by the
+  # server can be turned into a proper response.  Send back an error.
+  # Changed to 406 after considering rt.cpan.org 20975.
+  #
+  # 10.4.7 406 Not Acceptable
+  #
+  # The resource identified by the request is only capable of
+  # generating response entities which have content characteristics
+  # not acceptable according to the accept headers sent in the
+  # request.
+
+  $request->error(406, "Server response is Not Acceptable - $request_id");
 }
 
 # }}} _poco_weeble_io_error
@@ -562,14 +630,15 @@ sub _poco_weeble_io_read {
   DEBUG and warn "I/O: wheel $wheel_id got input...";
   DEBUG_DATA and warn (ref($input) ? $input->as_string : _hexdump($input));
 
-  # TODO - So, which is it?  Return, or die?
+  # There was no corresponding request?  Nothing left to do here.
+  # We might have got here because the server sent EOF after we were done processing
+  # the request, and deleted it from our cache. ( notes for RT#50231 )
   return unless defined $request_id;
-  die unless defined $request_id;
 
   my $request = $heap->{request}->{$request_id};
   return unless defined $request;
   DEBUG and warn(
-    "REQUEST $request_id is $request <" . $request->[REQ_REQUEST]->uri . ">"
+    "REQUEST $request_id is $request <" . $request->[REQ_HTTP_REQUEST]->uri . ">"
   );
 
   # Reset the timeout if we get data.
@@ -586,9 +655,10 @@ sub _poco_weeble_io_read {
   # part of the content.
   if ($request->[REQ_STATE] & RS_IN_HEAD) {
     if (defined $input) {
-      $input->request ($request->[REQ_REQUEST]);
+      $input->request ($request->[REQ_HTTP_REQUEST]);
       #warn(
-      #  "INPUT for ", $request->[REQ_REQUEST]->uri, " is \n",$input->as_string
+      #  "INPUT for ", $request->[REQ_HTTP_REQUEST]->uri,
+      #  " is \n",$input->as_string
       #)
     }
     else {
@@ -599,13 +669,33 @@ sub _poco_weeble_io_read {
     # FIXME: This happens when the response is HTTP/0.9 and doesn't
     # include a status line.  See t/53_response_parser.t.
     $request->[REQ_RESPONSE] = $input;
+    $input->header("X-PCCH-Peer", $request->[REQ_PEERNAME]);
+
+    # TODO CONNECT - If we've got the headers to a CONNECT request,
+    # then we can switch to the actual request.  This is like a faux
+    # redirect where the socket gets reused.
+    #
+    # 1. Switch the socket to SSL.
+    # 2. Switch the request from CONNECT mode to regular mode, using
+    #    the method proposed in PCCH::Request.
+    # 3. Send the original request via PCCH::Request->send_to_wheel().
+    #    This puts the client back into the RS_SENDING state.
+    # 4. Reset any data/state so it appears we never went through
+    #    CONNECT.
+    # 5. Make sure that PCC::Keepalive will discard the socket when
+    #    we're done with it.
+    # 6. Return.  The connection should proceed as normal.
+    #
+    # I think the normal handling for HTTP errors will cover the case
+    # of CONNECT failure.  If not, we can refine the implementation as
+    # needed.
 
     # Some responses are without content by definition
     # FIXME: #12363
-    #        Make sure we finish even when it isn't one of these,
-    #        but there is no content.
+    # Make sure we finish even when it isn't one of these, but there
+    # is no content.
     if (
-      $request->[REQ_REQUEST]->method eq 'HEAD'
+      $request->[REQ_HTTP_REQUEST]->method eq 'HEAD'
       or $input->code =~ /^(?:1|[23]04)/
       or (
         defined($input->content_length())
@@ -617,18 +707,35 @@ sub _poco_weeble_io_read {
         delete $heap->{wheel_to_request}->{$wheel_id};
         if (defined $old_request) {
           DEBUG and warn "I/O: removed request $request_id";
-          delete $heap->{request_to_id}{$old_request->[REQ_REQUEST]};
           $old_request->remove_timeout();
+          delete $heap->{ext_request_to_int_id}{$old_request->[REQ_HTTP_REQUEST]};
           $old_request->[REQ_CONNECTION] = undef;
         }
         return;
       }
       $request->[REQ_STATE] |= RS_DONE;
       $request->remove_timeout();
-      _finish_request($heap, $request, 1);
+      _finish_request($heap, $request);
       return;
     }
     else {
+      # If we have content length, and it's more than the maximum we
+      # requested, then fail without bothering with the content.
+      if (
+        defined($heap->{factory}->max_response_size())
+        and defined($input->content_length())
+        and $input->content_length() > $heap->{factory}->max_response_size()
+      ) {
+        _internal_cancel(
+          $heap, $request_id, 406,
+          "Response content length " . $input->content_length() .
+          " is greater than specified MaxSize of " .
+          $heap->{factory}->max_response_size() .
+          ".  Use range requests to retrieve specific amounts of content."
+        );
+        return;
+      }
+
       $request->[REQ_STATE] |= RS_IN_CONTENT;
       $request->[REQ_STATE] &= ~RS_IN_HEAD;
       #FIXME: probably want to find out when the content from this
@@ -639,7 +746,7 @@ sub _poco_weeble_io_read {
         delete $heap->{wheel_to_request}->{$wheel_id};
         if (defined $old_request) {
           DEBUG and warn "I/O: removed request $request_id";
-          delete $heap->{request_to_id}{$old_request->[REQ_REQUEST]};
+          delete $heap->{ext_request_to_int_id}{$old_request->[REQ_HTTP_REQUEST]};
           $old_request->remove_timeout();
           $old_request->[REQ_CONNECTION]->close();
           $old_request->[REQ_CONNECTION] = undef;
@@ -735,7 +842,7 @@ sub _poco_weeble_io_read {
     not $request->[REQ_STATE] & RS_POSTED
   ) {
     $request->remove_timeout;
-    _finish_request($heap, $request, 1);
+    _finish_request($heap, $request);
   }
 
 # }}} deliver reponse if complete
@@ -809,7 +916,7 @@ sub _try_redirect {
 # {{{ _finish_request
 
 sub _finish_request {
-  my ($heap, $request, $wait) = @_;
+  my ($heap, $request) = @_;
 
   my $request_id = $request->ID;
   if (DEBUG) {
@@ -833,31 +940,7 @@ sub _finish_request {
 
   DEBUG and warn "address is $address";
 
-  if ($wait) {
-    #wait a bit with removing the request, so there's
-    #time to receive the EOF event in case the connection
-    #gets closed.
-    my $alarm_id = $poe_kernel->delay_set('remove_request', 0.5, $request_id);
-
-    # remove the old timeout first
-    DEBUG and warn "delay_set; now remove_timeout()";
-    $request->remove_timeout();
-    DEBUG and warn "removed timeout; now timer()";
-    $request->timer($alarm_id);
-  }
-  else {
-    # Virtually identical to _remove_request.
-    # TODO - Make a common sub to handle both cases?
-    my $request = delete $heap->{request}->{$request_id};
-    if (defined $request) {
-      DEBUG and warn "I/O: removing request $request_id";
-      $request->remove_timeout();
-      if (my $wheel = $request->wheel) {
-        delete $heap->{wheel_to_request}->{$wheel->ID};
-        delete $heap->{request_to_id}{$request->[REQ_REQUEST]};
-      }
-    }
-  }
+  return _clear_req_cache( $heap, $request_id );
 }
 
 # }}} _finish_request
@@ -866,23 +949,34 @@ sub _finish_request {
 sub _poco_weeble_remove_request {
   my ($kernel, $heap, $request_id) = @_[KERNEL, HEAP, ARG0];
 
+  return _clear_req_cache( $heap, $request_id );
+}
+#}}} _remove_request
+
+# helper subroutine to remove a request from our caches
+#{{{ _clear_req_cache
+sub _clear_req_cache {
+  my ($heap, $request_id) = @_;
+
   my $request = delete $heap->{request}->{$request_id};
   if (defined $request) {
     DEBUG and warn "I/O: removed request $request_id";
     $request->remove_timeout();
+    delete $heap->{ext_request_to_int_id}{$request->[REQ_HTTP_REQUEST]};
     if (my $wheel = $request->wheel) {
       delete $heap->{wheel_to_request}->{$wheel->ID};
-      delete $heap->{request_to_id}{$request->[REQ_REQUEST]};
     }
   }
+
+  return;
 }
-#}}} _remove_request
+#}}} _clear_req_cache
 
 # Cancel a single request by HTTP::Request object.
 
 sub _poco_weeble_cancel {
   my ($kernel, $heap, $request) = @_[KERNEL, HEAP, ARG0];
-  my $request_id = $heap->{request_to_id}{$request};
+  my $request_id = $heap->{ext_request_to_int_id}{$request};
   return unless defined $request_id;
   _internal_cancel(
     $heap, $request_id, 408, "Request timed out (request canceled)"
@@ -895,14 +989,14 @@ sub _internal_cancel {
   my $request = delete $heap->{request}{$request_id};
   return unless defined $request;
 
-  DEBUG and warn "SHT: Shutdown is canceling request $request_id";
+  DEBUG and warn "CXL: canceling request $request_id";
   $request->remove_timeout();
+  delete $heap->{ext_request_to_int_id}{$request->[REQ_HTTP_REQUEST]};
 
   if (my $wheel = $request->wheel) {
     my $wheel_id = $wheel->ID;
-    DEBUG and warn "SHT: Request $request_id canceling wheel $wheel_id";
+    DEBUG and warn "CXL: Request $request_id canceling wheel $wheel_id";
     delete $heap->{wheel_to_request}{$wheel_id};
-    delete $heap->{request_to_id}{$request->[REQ_REQUEST]};
     $wheel = undef;
   }
 
@@ -917,7 +1011,7 @@ sub _internal_cancel {
   }
 
   unless ($request->[REQ_STATE] & RS_POSTED) {
-    $request->error(408, "Request timed out (component shut down)");
+    $request->error($code, $message);
   }
 }
 
@@ -936,7 +1030,7 @@ sub _poco_weeble_shutdown {
 
   # Shut down the connection manager subcomponent.
   if (defined $heap->{cm}) {
-    DEBUG and warn "SHT: Client::HTTP shutting down Client::Keepalive";
+    DEBUG and warn "CXL: Client::HTTP shutting down Client::Keepalive";
     $heap->{cm}->shutdown();
     delete $heap->{cm};
   }
@@ -955,6 +1049,10 @@ __END__
 
 POE::Component::Client::HTTP - a HTTP user-agent component
 
+=head1 VERSION
+
+version 0.943
+
 =head1 SYNOPSIS
 
   use POE qw(Component::Client::HTTP);
@@ -967,7 +1065,7 @@ POE::Component::Client::HTTP - a HTTP user-agent component
     Timeout   => 60,                    # defaults to 180 seconds
     MaxSize   => 16384,                 # defaults to entire response
     Streaming => 4096,                  # defaults to 0 (off)
-    FollowRedirects => 2                # defaults to 0 (off)
+    FollowRedirects => 2,               # defaults to 0 (off)
     Proxy     => "http://localhost:80", # defaults to HTTP_PROXY env. variable
     NoProxy   => [ "localhost", "127.0.0.1" ], # defs to NO_PROXY env. variable
     BindAddr  => "12.34.56.78",         # defaults to INADDR_ANY
@@ -1020,9 +1118,9 @@ POE::Component::Client::HTTP is an HTTP user-agent for POE.  It lets
 other sessions run while HTTP transactions are being processed, and it
 lets several HTTP transactions be processed in parallel.
 
-If POE::Component::Client::DNS is also installed, Client::HTTP will
-use it to resolve hosts without blocking.  Otherwise it will use
-gethostbyname(), which may have performance problems.
+It supports keep-alive through POE::Component::Client::Keepalive,
+which in turn uses POE::Component::Resolver for asynchronous IPv4 and
+IPv6 name resolution.
 
 HTTP client components are not proper objects.  Instead of being
 created, as most objects are, they are "spawned" as separate sessions.
@@ -1080,7 +1178,9 @@ not have done this already.
     # ...
   );
 
-See L<POE::Component::Client::Keepalive> for more information.
+See L<POE::Component::Client::Keepalive> for more information,
+including how to alter the connection manager's resolver
+configuration (for example, to force IPv6 or prefer it before IPv4).
 
 =item CookieJar => $cookie_jar
 
@@ -1188,8 +1288,48 @@ for details here.
 
 =item Timeout => $query_timeout
 
-C<Timeout> specifies the amount of time a HTTP request will wait for
-an answer.  This defaults to 180 seconds (three minutes).
+C<Timeout> sets how long POE::Component::Client::HTTP has to process
+an application's request, in seconds.  C<Timeout> defaults to 180
+(three minutes) if not specified.
+
+It's important to note that the timeout begins when the component
+receives an application's request, not when it attempts to connect to
+the web server.
+
+Timeouts may result from sending the component too many requests at
+once.  Each request would need to be received and tracked in order.
+Consider this:
+
+  $_[KERNEL]->post(component => request => ...) for (1..15_000);
+
+15,000 requests are queued together in one enormous bolus.  The
+component would receive and initialize them in order.  The first
+socket activity wouldn't arrive until the 15,000th request was set up.
+If that took longer than C<Timeout>, then the requests that have
+waited too long would fail.
+
+C<ConnectionManager>'s own timeout and concurrency limits also affect
+how many requests may be processed at once.  For example, most of the
+15,000 requests would wait in the connection manager's pool until
+sockets become available.  Meanwhile, the C<Timeout> would be counting
+down.
+
+Applications may elect to control concurrency outside the component's
+C<Timeout>.  They may do so in a few ways.
+
+The easiest way is to limit the initial number of requests to
+something more manageable.  As responses arrive, the application
+should handle them and start new requests.  This limits concurrency to
+the initial request count.
+
+An application may also outsource job throttling to another module,
+such as POE::Component::JobQueue.
+
+In any case, C<Timeout> and C<ConnectionManager> may be tuned to
+maximize timeouts and concurrency limits.  This may help in some
+cases.  Developers should be aware that doing so will increase memory
+usage.  POE::Component::Client::HTTP and KeepAlive track requests in
+memory, while applications are free to keep pending requests on disk.
 
 =back
 
@@ -1228,6 +1368,10 @@ must be invoked with $kernel->call().
 
   my $count = $kernel->call('ua' => 'pending_requests_count');
 
+NOTE: Sometimes the count might not be what you expected, because responses
+are currently in POE's queue and you haven't processed them. This could happen
+if you configure the C<ConnectionManager>'s concurrency to a high enough value.
+
 =head2 cancel
 
 Cancel a specific HTTP request.  Requires a reference to the original
@@ -1311,11 +1455,11 @@ each time it is called, and an empty string when done.  Don't forget
 to set the Content-Length header correctly.  Example:
 
   my $request = HTTP::Request->new( PUT => 'http://...' );
-  
+
   my $file = '/path/to/large_file';
-  
+
   open my $fh, '<', $file;
-  
+
   my $upload_cb = sub {
     if ( sysread $fh, my $buf, 4096 ) {
       return $buf;
@@ -1325,11 +1469,11 @@ to set the Content-Length header correctly.  Example:
       return '';
     }
   };
-  
+
   $request->content_length( -s $file );
-  
+
   $request->content( $upload_cb );
-  
+
   $kernel->post( ua => request, 'response', $request );
 
 =head1 CONTENT ENCODING AND COMPRESSION
@@ -1348,7 +1492,7 @@ use HTTP::Response's decoded_content() method rather than content():
   );
 
   # ... time passes ...
-  
+
   my $content = $response->decoded_content();
 
 The change in POE::Component::Client::HTTP behavior was prompted by
@@ -1363,6 +1507,33 @@ solution, see:
 L<http://www.perlmonks.org/?node_id=683833> and
 L<http://rt.cpan.org/Ticket/Display.html?id=35538>
 
+=head1 CLIENT HEADERS
+
+POE::Component::Client::HTTP sets its own response headers with
+additional information.  All of its headers begin with "X-PCCH".
+
+=head2 X-PCCH-Errmsg
+
+POE::Component::Client::HTTP may fail because of an internal client
+error rather than an HTTP protocol error.  X-PCCH-Errmsg will contain a
+human readable reason for client failures, should they occur.
+
+The text of X-PCCH-Errmsg may also be repeated in the response's
+content.
+
+=head2 X-PCCH-Peer
+
+X-PCCH-Peer contains the remote IPv4 address and port, separated by a
+period.  For example, "127.0.0.1.8675" represents port 8675 on
+localhost.
+
+Proxying will render X-PCCH-Peer nearly useless, since the socket will
+be connected to a proxy rather than the server itself.
+
+This feature was added at Doreen Grey's request.  Doreen wanted a
+means to find the remote server's address without having to make an
+additional request.
+
 =head1 ENVIRONMENT
 
 POE::Component::Client::HTTP uses two standard environment variables:
@@ -1393,9 +1564,9 @@ Secure HTTP (https) proxying is not supported at this time.
 
 There is no object oriented interface.  See
 L<POE::Component::Client::Keepalive> and
-L<POE::Component::Client::DNS> for examples of a decent OO interface.
+L<POE::Component::Resolver> for examples of a decent OO interface.
 
-=head1 AUTHOR, COPYRIGHT, & LICENSE 
+=head1 AUTHOR, COPYRIGHT, & LICENSE
 
 POE::Component::Client::HTTP is
 
@@ -1403,7 +1574,7 @@ POE::Component::Client::HTTP is
 
 =item
 
-Copyright 1999-2006 Rocco Caputo
+Copyright 1999-2009 Rocco Caputo
 
 =item
 
@@ -1427,17 +1598,19 @@ L<http://www.sapo.pt/> was kind enough to support his contributions.
 Jeff Bisbee added POD tests and documentation to pass several of them
 to version 0.79.  He's a kwalitee-increasing machine!
 
-=head1 CONTACT
+=head1 BUG TRACKER
 
-Rocco may be contacted by e-mail via L<mailto:rcaputo@cpan.org>, and
-Martijn may be contacted by email via L<mailto:martijn@cpan.org>.
+https://rt.cpan.org/Dist/Display.html?Queue=POE-Component-Client-HTTP
 
-The preferred way to report bugs or requests is through RT though.
-See L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=POE-Component-Client-HTTP>
-or mail L<mailto:bug-POE-Component-Client-HTTP@rt.cpan.org>
+=head1 REPOSITORY
 
-For questions, try the L<POE> mailing list (poe@perl.org)
+http://github.com/rcaputo/poe-component-client-http
+http://gitorious.org/poe-component-client-http
+
+=head1 OTHER RESOURCES
+
+http://search.cpan.org/dist/POE-Component-Client-HTTP/
 
 =cut
 
-# }}} POD
+# }}} POD
\ No newline at end of file
